
#include <vector>
#include <set>
#include <string>
#include <sstream>
#include <algorithm>
#include <stdexcept>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <ctime>
#include <cstdarg>

using namespace std;


////////////////////////////////////////////////////////////
// [0, 1] uniformly distributed random numbers
//
double uniform()
{
	double r = (rand() + 0.5) / (RAND_MAX + 1.0);
	assert(r >= 0.0 && r <= 1.0);
	return r;
}

////////////////////////////////////////////////////////////
// descriptive statistics
//

double mean(std::vector<double> const& data)
{
	double sum = 0.0;
	for (size_t i=0; i<data.size(); i++) sum += data[i];
	return (sum / data.size());
}

double stddev(std::vector<double> const& data)
{
	double m = mean(data);
	double sum = 0.0;
	for (size_t i=0; i<data.size(); i++)
	{
		double diff = data[i] - m;
		sum += diff * diff;
	}
	return sqrt(sum / data.size());
}

double quantile(std::vector<double> data, double q)
{
	assert(q >= 0.0 && q <= 1.0);
	std::sort(data.begin(), data.end());
	double index = q * (data.size() - 1.0);
	double remainder = index - floor(index);
	size_t i = (size_t)floor(index);
	assert(i < data.size());
	double v = data[i];
	double w = (i + 1 < data.size()) ? data[i + 1] : data[i];
	return ((1.0 - remainder) * v + remainder * w);
}

////////////////////////////////////////////////////////////
// simplistic vectors and matrices
//
template <size_t DIM>
struct Vector
{
	Vector()
	{ }

	// specialized constructor for DIM=1 and DIM=2
	Vector(double x, double y = 0.0)
	{
		data[0] = x;
		if (DIM >= 2) data[1] = y;
	}

	size_t size() const
	{ return DIM; }
	double& operator [] (size_t index)
	{ return data[index]; }
	double const& operator [] (size_t index) const
	{ return data[index]; }

	void operator += (Vector<DIM> const& rhs)
	{ for (size_t i=0; i<DIM; i++) data[i] += rhs[i]; }
	Vector<DIM> operator + (Vector<DIM> const& rhs) const
	{ Vector<DIM> ret = *this; ret += rhs; return ret; }

	double data[DIM];
};

template <size_t ROWS, size_t COLS>
struct Matrix
{
	Matrix()
	{ }

	Matrix(double val)
	{
		for (size_t i=0; i<ROWS; i++) for (size_t j=0; j<COLS; j++) value[i][j] = val;
	}

	double& operator () (size_t i, size_t j)
	{ return value[i][j]; }
	double const& operator () (size_t i, size_t j) const
	{ return value[i][j]; }

	double value[ROWS][COLS];
};

template <size_t DIM>
double operator * (Vector<DIM> const& lhs, Vector<DIM> const& rhs)
{
	double ret = 0.0;
	for (size_t i=0; i<DIM; i++) ret += lhs[i] * rhs[i];
	return ret;
}

template <size_t DIM>
Vector<DIM> operator * (double lhs, Vector<DIM> const& rhs)
{
	Vector<DIM> ret;
	for (size_t i=0; i<DIM; i++) ret[i] = lhs * rhs[i];
	return ret;
}

template <size_t INDIM, size_t OUTDIM>
Vector<OUTDIM> operator * (Vector<INDIM> const& lhs, Matrix<INDIM, OUTDIM> const& rhs)
{
	Vector<OUTDIM> ret;
	for (size_t i=0; i<OUTDIM; i++)
	{
		double sum = 0.0;
		for (size_t j=0; j<INDIM; j++) sum += lhs[j] * rhs(j, i);
		ret[i] = sum;
	}
	return ret;
}

typedef Vector<1> Vector1;
typedef Vector<2> Vector2;
typedef Matrix<2, 1> Matrix2x1;

////////////////////////////////////////////////////////////
// dominance relation
//
bool dominates(Vector2 const& lhs, Vector2 const& rhs)
{ return ((lhs[0] <= rhs[0]) && (lhs[1] <= rhs[1])); }

////////////////////////////////////////////////////////////
// comparator for sorting by last objective
//
struct ComparatorIndicesLast
{
	ComparatorIndicesLast(const vector<Vector2>* comp_points_)
	: comp_points(comp_points_)
	{ }

	const vector<Vector2>* comp_points;

	bool operator () (size_t i, size_t j)
	{
		Vector2 const& vi = (*comp_points)[i];
		Vector2 const& vj = (*comp_points)[j];
		if (vi[1] == vj[1]) return (vi[0] < vj[0]);
		else return (vi[1] < vj[1]);
	}
};

////////////////////////////////////////////////////////////
// representation of a cuboid as a set of six point indices
//
struct Rectangle
{
	size_t a;     // lower bound x
	size_t b;     // upper bound x
	size_t c;     // lower bound y
	size_t d;     // upper bound y
};

////////////////////////////////////////////////////////////
// 2D hypervolume computation in N log(N) operations.
// The algorithm decomposes the dominated hypervolume
// into a disjoint union of rectangles, represented
// as a sequence of point indices (which thus allows for
// differentiation, etc.). Each 4-tuple of indices
// (a, b, c, d) represents a rectangle area
//    (x_b - x_a) * (y_d - y_c)
// where (x, y, z) are point coordinates. The reference
// point is represented by index value N, which is the
// number of points.
//
double hypervolume(Vector2 const& reference, vector<Vector2> const& points, vector<Rectangle>& cuboids)
{
	cuboids.clear();
	size_t N = points.size();
	if (N == 0) return 0.0;

#ifdef DEBUG
	for (size_t i=0; i<N; i++) assert(dominates(points[i], reference));
#endif

	// sort indices by last objective
	ComparatorIndicesLast comparator_last(&points);
	vector<size_t> index(N);
	for (size_t i=0; i<N; i++) index[i] = i;
	sort(index.begin(), index.end(), comparator_last);

	// create cuboids representation and compute hypervolume
	cuboids.resize(N);
	Rectangle rect;
	rect.a = index[0];
	rect.b = N;
	rect.c = index[0];
	rect.d = N;
	cuboids[0] = rect;
	double hv = (reference[0] - points[rect.a][0]) * (reference[1] - points[rect.c][1]);
	for (size_t i=1; i<N; i++)
	{
		assert(! dominates(points[index[i-1]], points[index[i]]));
		rect.a = index[i];
		rect.b = index[i-1];
		rect.c = index[i];
		rect.d = N;
		cuboids[i] = rect;
		hv += (points[rect.b][0] - points[rect.a][0]) * (reference[1] - points[rect.c][1]);
	}

	return hv;
}

////////////////////////////////////////////////////////////
// expressions
//

struct Expression
{
	virtual ~Expression()
	{ }

	virtual string str() const = 0;
	virtual double eval(double u) const = 0;
	virtual double deriv(double u, double& der) const = 0;
};

struct BinaryOperator : public Expression
{
	BinaryOperator(Expression* l, Expression* r, char sym)
	: lhs(l)
	, rhs(r)
	, symbol(sym)
	{ }

	~BinaryOperator()
	{ delete lhs; delete rhs; }

	string str() const
	{ return ("(" + lhs->str() + symbol + rhs->str() + ")"); }

	Expression* lhs;
	Expression* rhs;
	char symbol;
};

struct Sum : public BinaryOperator
{
	Sum(Expression* l, Expression* r)
	: BinaryOperator(l, r, '+')
	{ }

	double eval(double u) const
	{ return (lhs->eval(u) + rhs->eval(u)); }
	double deriv(double u, double& der) const
	{
		double der2;
		double ret;
		ret = lhs->deriv(u, der);
		ret += rhs->deriv(u, der2);
		der += der2;
		return ret;
	}
};

struct Difference : public BinaryOperator
{
	Difference(Expression* l, Expression* r)
	: BinaryOperator(l, r, '-')
	{ }

	double eval(double u) const
	{ return (lhs->eval(u) - rhs->eval(u)); }
	double deriv(double u, double& der) const
	{
		double der2;
		double ret;
		ret = lhs->deriv(u, der);
		ret -= rhs->deriv(u, der2);
		der -= der2;
		return ret;
	}
};

struct Product : public BinaryOperator
{
	Product(Expression* l, Expression* r)
	: BinaryOperator(l, r, '*')
	{ }

	double eval(double u) const
	{ return (lhs->eval(u) * rhs->eval(u)); }
	double deriv(double u, double& der) const
	{
		double der1, der2;
		double val1 = lhs->deriv(u, der1);
		double val2 = rhs->deriv(u, der2);
		der = val2 * der1 + val1 * der2;
		return (val1 * val2);
	}
};

struct Quotient : public BinaryOperator
{
	Quotient(Expression* l, Expression* r)
	: BinaryOperator(l, r, '/')
	{ }

	double eval(double u) const
	{ return (lhs->eval(u) / rhs->eval(u)); }
	double deriv(double u, double& der) const
	{
		double der1, der2;
		double val1 = lhs->deriv(u, der1);
		double val2 = rhs->deriv(u, der2);
		der = (val2 * der1 - val1 * der2) / (val2 * val2);
		return (val1 / val2);
	}
};

struct Power : public BinaryOperator
{
	Power(Expression* l, Expression* r)
	: BinaryOperator(l, r, '^')
	{ }

	double eval(double u) const
	{ return (pow(lhs->eval(u), rhs->eval(u))); }
	double deriv(double u, double& der) const
	{
		double der1, der2;
		double val1 = lhs->deriv(u, der1);
		double val2 = rhs->deriv(u, der2);
		double ret = pow(val1, val2);
		double d1 = ret * val2 / val1;
		double d2 = ret * log(val1);
		der = d1 * der1 + d2 * der2;
		return ret;
	}
};

struct Negation : public Expression
{
	Negation(Expression* e)
	: ex(e)
	{ }

	~Negation()
	{ delete ex; }

	string str() const
	{ return ("-" + ex->str()); }
	double eval(double u) const
	{ return (-ex->eval(u)); }
	double deriv(double u, double& der) const
	{
		double ret = ex->deriv(u, der);
		der = -der;
		return ret;
	}

	Expression* ex;
};

struct Constant : public Expression
{
	Constant(double v)
	: value(v)
	{ }

	string str() const
	{ stringstream ss; ss << value; return ss.str(); }
	double eval(double u) const
	{ return value; }
	double deriv(double u, double& der) const
	{ der = 0.0; return value; }

	double value;
};

struct NamedConstant : public Constant
{
	NamedConstant(string n, double v)
	: Constant(v)
	, name(n)
	{ }

	string str() const
	{ return name; }

	string name;
};

struct Variable : public Expression
{
	Variable()
	{ }

	string str() const
	{ return "u"; }
	double eval(double u) const
	{ return u; }
	double deriv(double u, double& der) const
	{
		der = 1.0;
		return u;
	}
};

typedef double(*func)(double);

struct Function : public Expression
{
	Function(string name_, func f_, func df_, Expression* inner_)
	: name(name_)
	, f(f_)
	, df(df_)
	, inner(inner_)
	{ }

	~Function()
	{ delete inner; }

	string str() const
	{ return (name + "(" + inner->str() + ")"); }
	double eval(double u) const
	{ return f(inner->eval(u)); }
	double deriv(double u, double& der) const
	{
		double i = inner->deriv(u, der);
		double d = df(i);
		der *= d;
		return f(i);
	}

	string name;
	func f;
	func df;
	Expression* inner;
};

struct Switch : public Expression
{
	Switch(vector<Expression*> const& ex, vector<double> const& thres)
	: inner(ex)
	, threshold(thres)
	{ assert(thres.size() + 1 == inner.size()); }

	~Switch()
	{ for (size_t i=0; i<inner.size(); i++) delete inner[i]; }

	string str() const
	{
		stringstream ss;
		ss << "switch(";
		for (size_t i=0; i<threshold.size(); i++) ss << inner[i]->str() << "; " << threshold[i] << "; ";
		ss << inner[threshold.size()]->str() << ")";
		return ss.str();
	}
	double eval(double u) const
	{
		size_t i=0;
		for (i=0; i<threshold.size(); i++) if (u < threshold[i]) break;
		return inner[i]->eval(u);
	}
	double deriv(double u, double& der) const
	{
		size_t i=0;
		for (i=0; i<threshold.size(); i++) if (u < threshold[i]) break;
		return inner[i]->deriv(u, der);
	}

	vector<Expression*> inner;
	vector<double> threshold;
};

double d_sqrt(double t)
{ return (0.5 / sqrt(t)); }

double d_log(double t)
{ return (1.0 / t); }

double d_cos(double t)
{ return (-sin(t)); }

double d_tan(double t)
{ t = tan(t); return (1.0 + t * t); }

double d_tanh(double t)
{ t = tanh(t); return (1.0 - t * t); }

double d_asin(double t)
{ return (1.0 / sqrt(1.0 - t * t)); }

double d_acos(double t)
{ return (-1.0 / sqrt(1.0 - t * t)); }

double d_atan(double t)
{ return (1.0 / (1.0 + t * t)); }

Expression* parse_composite(const char* str, size_t& position);

Expression* parse_simple(const char* str, size_t& position)
{
	while (str[position] == ' ') position++;
	char c = str[position];
	if (c == '(')
	{
		position++;
		Expression* ex = parse_composite(str, position);
		if (ex == NULL) return NULL;
		while (str[position] == ' ') position++;
		if (str[position] != ')')
		{
			delete ex;
			return NULL;
		}
		position++;
		return ex;
	}
	else if (c == '-')
	{
		position++;
		Expression* sub = parse_simple(str, position);
		if (sub == NULL) return NULL;
		return new Negation(sub);
	}
	else if ((c >= '0' && c <= '9') || c == '.')
	{
		// parse numeric constant
		char* endp;
		double value = strtod(str + position, &endp);
		if (endp == str + position) return NULL;
		position = endp - str;
		return new Constant(value);
	}
	else if (c >= 'a' && c <= 'z')
	{
		// parse token
		size_t start = position;
		position++;
		while (str[position] >= 'a' && str[position] <= 'z') position++;
		string token(str + start, position - start);
		if (token == "e")
		{
			return new NamedConstant("e", M_E);
		}
		else if (token == "pi")
		{
			return new NamedConstant("pi", M_PI);
		}
		else if (token == "u")
		{
			return new Variable();
		}
		else if (token == "switch")
		{
			while (str[position] == ' ') position++;
			if (str[position] != '(') return NULL;
			position++;

			vector<Expression*> inner;
			vector<double> threshold;
			Expression* ex = parse_composite(str, position);
			if (ex == NULL) return NULL;
			inner.push_back(ex);

			while (true)
			{
				while (str[position] == ' ') position++;
				if (str[position] == ')')
				{
					position++;
					return new Switch(inner, threshold);
				}
				else if (str[position] == ';') position++;
				else return NULL;

				while (str[position] == ' ') position++;
				char* end = NULL;
				double t = strtod(str + position, &end);
				if (end == str) return NULL;
				threshold.push_back(t);
				position = end - str;

				while (str[position] == ' ') position++;
				if (str[position] != ';') return NULL;
				position++;

				Expression* ex = parse_composite(str, position);
				if (ex == NULL) return NULL;
				inner.push_back(ex);
			}
		}
		else
		{
			// assume a function
			while (str[position] == ' ') position++;
			if (str[position] != '(') return NULL;
			position++;
			Expression* sub = parse_composite(str, position);
			if (sub == NULL) return NULL;
			while (str[position] == ' ') position++;
			if (str[position] != ')') { delete sub; return NULL; }
			position++;

			if (token == "sqrt") return new Function(token, sqrt, d_sqrt, sub);
			else if (token == "exp") return new Function(token, exp, exp, sub);
			else if (token == "log") return new Function(token, log, d_log, sub);
			else if (token == "sin") return new Function(token, sin, cos, sub);
			else if (token == "cos") return new Function(token, cos, d_cos, sub);
			else if (token == "tan") return new Function(token, tan, d_tan, sub);
			else if (token == "sinh") return new Function(token, sinh, cosh, sub);
			else if (token == "cosh") return new Function(token, cosh, sinh, sub);
			else if (token == "tanh") return new Function(token, tanh, d_tan, sub);
			else if (token == "asin") return new Function(token, asin, d_asin, sub);
			else if (token == "acos") return new Function(token, acos, d_acos, sub);
			else if (token == "atan") return new Function(token, atan, d_atan, sub);
			else
			{
				delete sub;
				return NULL;
			}
		}
	}
	return NULL;
}

Expression* parse_composite(const char* str, size_t& position)
{
	vector<Expression*> sub;
	vector<char> symbol;
	string symbols = "+-*/^";
	while (true)
	{
		Expression* ex = parse_simple(str, position);
		if (ex == NULL)
		{
			for (size_t i=0; i<sub.size(); i++) delete sub[i];
			return NULL;
		}
		sub.push_back(ex);

		while (str[position] == ' ') position++;
		char c = str[position];
		if (c == ')' || c == 0) break;
		if (symbols.find(c) == string::npos) break;
		symbol.push_back(c);
		position++;
	}

	// find ^ operators
	for (size_t i=0; i<symbol.size(); i++)
	{
		if (symbol[i] == '^')
		{
			Expression* lhs = sub[i];
			Expression* rhs = sub[i+1];
			Expression* ex = new Power(lhs, rhs);
			sub[i] = ex;
			sub.erase(sub.begin() + i + 1);
			symbol.erase(symbol.begin() + i);
			i--;
		}
	}

	// find * and / operators
	for (size_t i=0; i<symbol.size(); i++)
	{
		if (symbol[i] == '*')
		{
			Expression* lhs = sub[i];
			Expression* rhs = sub[i+1];
			Expression* ex = new Product(lhs, rhs);
			sub[i] = ex;
			sub.erase(sub.begin() + i + 1);
			symbol.erase(symbol.begin() + i);
			i--;
		}
		else if (symbol[i] == '/')
		{
			Expression* lhs = sub[i];
			Expression* rhs = sub[i+1];
			Expression* ex = new Quotient(lhs, rhs);
			sub[i] = ex;
			sub.erase(sub.begin() + i + 1);
			symbol.erase(symbol.begin() + i);
			i--;
		}
	}

	// find + and - operators
	for (size_t i=0; i<symbol.size(); i++)
	{
		if (symbol[i] == '+')
		{
			Expression* lhs = sub[i];
			Expression* rhs = sub[i+1];
			Expression* ex = new Sum(lhs, rhs);
			sub[i] = ex;
			sub.erase(sub.begin() + i + 1);
			symbol.erase(symbol.begin() + i);
			i--;
		}
		else if (symbol[i] == '-')
		{
			Expression* lhs = sub[i];
			Expression* rhs = sub[i+1];
			Expression* ex = new Difference(lhs, rhs);
			sub[i] = ex;
			sub.erase(sub.begin() + i + 1);
			symbol.erase(symbol.begin() + i);
			i--;
		}
	}

	if (sub.size() != 1 || symbol.size() != 0)
	{
		for (size_t i=0; i<sub.size(); i++) delete sub[i];
		return NULL;
	}
	return sub[0];
}

Expression* parse(const char* str)
{
	size_t position = 0;
	Expression* ret = parse_composite(str, position);
	while (str[position] == ' ') position++;
	if (str[position] != 0) { delete ret; ret = NULL; }
	if (ret == NULL)
	{
		string s = str;
		cout << "parse error in expression '" << s << "' at position " << position << ":" << endl;
		cout << "      " << s.substr(0, position) << endl;
		cout << "      " << string(position, ' ') << s.substr(position) << endl;
	}
	return ret;
}

// parameterization \phi : [0, 1] \to \R^2 of the Pareto front
struct Phi
{
	Phi(Expression* phi0, Expression* phi1)
	{
		ex[0] = phi0;
		ex[1] = phi1;
	}


	Vector2 operator () (Vector1 const& param) const
	{ return eval(param); }

	Vector2 eval(Vector1 const& param) const
	{
		double u = param[0];
		return Vector2(
				ex[0]->eval(u),
				ex[1]->eval(u)
			);
	}

	Vector2 deriv(Vector1 const& param, Matrix2x1& der) const
	{
		Vector2 ret;
		double u = param[0];
		double d;
		ret[0] = ex[0]->deriv(u, d);
		der(0, 0) = d;
		ret[1] = ex[1]->deriv(u, d);
		der(1, 0) = d;
		return ret;
	}

	Expression* ex[2];
};

////////////////////////////////////////////////////////////
// unnormalized approximate parameter density
//
double density(Vector1 const& param, Phi const& phi)
{
	// compute front normal vector
	Matrix2x1 der;
	phi.deriv(param, der);
	Vector2 n;
	n[0] = der(1, 0);
	n[1] = -der(0, 0);
	double u_length = sqrt(n[0] * n[0] + n[1] * n[1]);

	// compute unit rectangle
	double c = sqrt(n[0] * n[1]);
	Vector2 b(c / n[0], c / n[1]);

	// compute length
	double length = sqrt(b[0] * b[0] + b[1] * b[1]);

	// compute density
	return (u_length / length);
}


////////////////////////////////////////////////////////////
// console help
//
void help()
{
	cout << endl;
	cout << "2D hypervolume maximizer (version 03/2014)" << endl;
	cout << "usage: hyp2d <f1> <f2> <ref1> <ref2> <N> [options]" << endl;
	cout << "with parameters" << endl;
	cout << "   f1                first component of the parameterization" << endl;
	cout << "   f2                second component of the parameterization" << endl;
	cout << "   ref1              number; first coordinate of the reference point" << endl;
	cout << "   ref2              number; second coordinate of the reference point" << endl;
	cout << "   n                 integer; size of the set" << endl;
	cout << "and options" << endl;
	cout << "   -repeat <number>  number of restarts of the gradient-based optimization" << endl;
	cout << "   -accuracy <real>  relative solution accuracy, default 1e-12" << endl;
	cout << "   -csv <filename>   point set (front) is stored in csv file" << endl;
	cout << "   -eps <filename>   set is exported to encapsulated postscript" << endl;
	cout << endl;
	cout << "The program generates a set of n points, maximizing the dominated hypervolume." << endl;
	cout << "The Pareto front is specified as a parametric surface of parameter u in" << endl;
	cout << "the unit interval [0, 1], given by formulas f1 and f2. The formulas may" << endl;
	cout << "be composed of numeric constants, symbol u, binary operators + - * /" << endl;
	cout << "and ^ (power), left unary operator -, brackets, and functions sqrt exp log" << endl;
	cout << "sin cos tan asin acos atan sinh cosh tanh, and named constants pi and e." << endl;
	cout << "Case distinction: switch(ex_1; c_1; ex_2; c_2; ...; ex_k)" << endl;
	cout << "If the argument is smaller than the i-th constant (c_i) but larger than the" << endl;
	cout << "(i-1)-th constant then the i-th expression ex_i is evaluated. This allows for" << endl;
	cout << "piecewise function definitions. Example (absolute value): switch(-u; 0; u)" << endl;
	cout << "The position of the reference point is (ref1, ref2)." << endl;
	cout << "The program outputs the hypervolume covered by the best set found as well as" << endl;
	cout << "distribution statistics for multiple optimization runs." << endl;
	cout << "NOTE: The Pareto front must be encoded such that formulas are valid for all" << endl;
	cout << "      u in [0, 1]. All points should be non-dominated." << endl;
	cout << "NOTE: Formulas with operators should be enclosed in quotes when typed on the" << endl;
	cout << "      shell." << endl;
	cout << endl;
	cout << "examples:" << endl;
	cout << endl;
	cout << "hyp2d '0.5*u' '0.5*(1-u)' 11 11 100" << endl;
	cout << endl;
	cout << "computes the maximal hypervolume that can be covered by 100 points on the unit" << endl;
	cout << "simplex w.r.t. the reference point (11, 11). This is the Pareto front of the" << endl;
	cout << "DTLZ1 problem. The result should be 120.87373737... ." << endl;
	cout << endl;
	cout << "hyp2d 'cos(pi*u/2)' 'sin(pi*u/2)' 11 11 100" << endl;
	cout << endl;
	cout << "performs a similar computation for the Pareto front of problems DTLZ2, DTLZ3," << endl;
	cout << "and DTLZ4." << endl;
	cout << endl;
	cout << "hyp2d 'u' '(1-u)*(1-u)' 11 11 100                                ZDT1 and ZDT4" << endl;
	cout << "hyp2d 'u' '1-u*u' 11 11 100                                               ZDT2" << endl;
	cout << "hyp2d '0.7192246809*u+0.2807753191' '1-(0.7192246809*u+0.2807753191)^2' \\" << endl;
	cout << "                                                 11 11 100                ZDT6" << endl;
	cout << endl;
	cout << "The following complex example is a parameterization of the Pareto front of the" << endl;
	cout << "ZDT3 problem:" << endl;
	cout << "hyp2d \"switch( 0.4150076745*u ; 0.2 ; 0.1822287280+0.377668177*(u-0.2); 0.4 \\" << endl;
	cout << "      ; 0.4093136748+0.2228421465*(u-0.4) ; 0.6 ; 0.6183967944+0.170574547* \\" << endl;
	cout << "      (u-0.6) ; 0.8 ; 0.8233317983+0.1425053355*(u-0.8) )\" \"1 - \\" << endl;
	cout << "      sqrt(switch( 0.4150076745*u ; 0.2 ; 0.1822287280+0.377668177*(u-0.2); \\" << endl;
	cout << "      0.4 ; 0.4093136748+0.2228421465*(u-0.4) ; 0.6 ; 0.6183967944+ \\" << endl;
	cout << "      0.170574547*(u-0.6) ; 0.8 ; 0.8233317983+0.1425053355*(u-0.8) )) - \\" << endl;
	cout << "      switch( 0.4150076745*u ; 0.2 ; 0.1822287280+0.377668177*(u-0.2); 0.4 ; \\" << endl;
	cout << "      0.4093136748+0.2228421465*(u-0.4) ; 0.6 ; 0.6183967944+0.170574547* \\" << endl;
	cout << "      (u-0.6) ; 0.8 ; 0.8233317983+0.1425053355*(u-0.8) ) * sin(10*pi* \\" << endl;
	cout << "      switch( 0.4150076745*u ; 0.2 ; 0.1822287280+0.377668177*(u-0.2); 0.4 ; \\" << endl;
	cout << "      0.4093136748+0.2228421465*(u-0.4) ; 0.6 ; 0.6183967944+0.170574547* \\" << endl;
	cout << "      (u-0.6) ; 0.8 ; 0.8233317983+0.1425053355*(u-0.8) ))\" 11 11 100" << endl;
	cout << endl;
	cout << "Append '-csv front.csv' (without quotes) to store the front to a csv file." << endl;
	cout << "Append '-eps front.eps' (without quotes) to visualize the front." << endl;
	cout << endl;

	exit(EXIT_FAILURE);
}

////////////////////////////////////////////////////////////
// entry point
//
int main(int argc, char** argv)
{
	// seed RNG
	srand(time(NULL));
//	srand(42);   // enable this if you need reproducible results

	size_t restarts = 1;
	size_t delta = 100;
	double accuracy = 1e-12;
	size_t maxiter = 100000;

	// parse command line
	if (argc < 6) help();
	Expression* ex0 = parse(argv[1]);
	if (ex0 == NULL) return EXIT_FAILURE;
	Expression* ex1 = parse(argv[2]);
	if (ex1 == NULL) return EXIT_FAILURE;
	Phi phi(ex0, ex1);
	Vector2 reference(strtod(argv[3], NULL), strtod(argv[4], NULL));
	size_t N = strtol(argv[5], NULL, 10);
	if (N == 0) help();
	string csv_filename, eps_filename;
	for (int a=6; a<argc; a++)
	{
		string arg = argv[a];
		if (arg == "-repeat")
		{
			a++;
			if (a < argc) restarts = strtol(argv[a], NULL, 10);
			else cout << "WARNING: no number of restarts given" << endl;
		}
		else if (arg == "-accuracy")
		{
			a++;
			if (a < argc) accuracy = strtod(argv[a], NULL);
			else cout << "WARNING: no accuracy given" << endl;
		}
		else if (arg == "-csv")
		{
			a++;
			if (a < argc) csv_filename = argv[a];
			else cout << "WARNING: no filename given" << endl;
		}
		else if (arg == "-eps")
		{
			a++;
			if (a < argc) eps_filename = argv[a];
			else cout << "WARNING: no filename given" << endl;
		}
		else cout << "WARNING: unrecognized parameter '" << arg << "'" << endl;
	}

	// dump parameters
	cout << "number of points: " << N << endl;
	cout << "reference point: (" << reference[0] << ", " << reference[1] << ")" << endl;
	cout << "parameterization:" << endl;
	cout << "    f1 = " << ex0->str() << endl;
	cout << "    f2 = " << ex1->str() << endl;
	cout << "repeats: " << restarts << endl;
	if (! csv_filename.empty()) cout << "set file: " << csv_filename << endl;
	if (! eps_filename.empty()) cout << "eps file: " << eps_filename << endl;

	// restart loop
	vector<double> hypervolume_distribution(restarts);
	double best_value = 0.0;
	vector<Vector2> best_points(N);
	for (size_t outer=0; outer<restarts; outer++)
	{
		// estimate upper bound on density
		double density_bound = 0.0;
		for (size_t i=0; i<100; i++)
		{
			Vector1 p(uniform());
			double d = density(p, phi);
			if (d > density_bound) density_bound = d;
		}
		density_bound *= 2.0;

		// initialize set with rejection sampling
		vector<Vector1> params(N);
		vector<Vector2> points(N);
		for (size_t i=0; i<N; )
		{
			params[i][0] = uniform();
			double d = density(params[i], phi);
			double r = uniform();
			if (r >= d / density_bound) continue;
			points[i] = phi(params[i]);
			i++;
		}

		// iteratively refine the set until convergence
		Vector2 zero2(0.0, 0.0);
		vector<Vector2> der_xy(N + 1);
		vector<Vector1> der_u(N);
		double eta = 1e-2;
		double old_h = 0.0;
		double prev_h = 0.0;
		vector<Vector1> backup = params;

		for (size_t iter=0; ; iter++)
		{
			// compute hypervolume
			vector<Rectangle> rectangles;
			double h = hypervolume(reference, points, rectangles);

			// check stopping criteria
			if ((iter % delta) == 0)
			{
				if (h - old_h <= abs(h) * accuracy || iter >= maxiter)
				{
					cout << "." << flush;
					hypervolume_distribution[outer] = h;
					if (h > best_value)
					{
						best_value = h;
						best_points = points;
					}
					break;
				}
				old_h = h;
			}
			if (! isfinite(h)) { cout << "!" << flush; break; }

			// learning rate adaptation
			if (h <= prev_h && eta > 1e-8)
			{
				// simple line search behavior
				for (size_t i=0; i<N; i++)
				{
					params[i] = 0.5 * (params[i] + backup[i]);
					points[i] = phi(params[i]);
				}
				eta *= 0.5;
				continue;
			}
			prev_h = h;
			eta *= 1.05;

			// compute derivatives w.r.t. points
			points.push_back(reference);
			for (size_t i=0; i<=N; i++) der_xy[i] = zero2;
			for (size_t i=0; i<rectangles.size(); i++)
			{
				Rectangle const& rect = rectangles[i];
				double dx = points[rect.b][0] - points[rect.a][0];
				double dy = points[rect.d][1] - points[rect.c][1];
				der_xy[rect.a][0] -= dy;
				der_xy[rect.b][0] += dy;
				der_xy[rect.c][1] -= dx;
				der_xy[rect.d][1] += dx;
			}
			points.pop_back();

			// compute derivative w.r.t. parameters
			for (size_t i=0; i<N; i++)
			{
				Matrix2x1 der;
				phi.deriv(params[i], der);
				der_u[i] = der_xy[i] * der;
			}

			// perform a gradient ascent step
			for (size_t i=0; i<N; i++)
			{
				backup[i] = params[i];
				params[i] += eta * der_u[i];
				if (params[i][0] <= 0.0) params[i][0] = 0.0;
				else if (params[i][0] >= 1.0) params[i][0] = 1.0;
				points[i] = phi(params[i]);
			}
		}
	}
	cout << endl;
	cout << "largest dominated hypervolume: " << setprecision(10) << best_value << endl;
	cout << "mean: " << mean(hypervolume_distribution) << endl;
	cout << "stddev: " << stddev(hypervolume_distribution) << endl;
	cout << "quantiles" << endl;
	cout << "   25%: " << quantile(hypervolume_distribution, 0.25) << endl;
	cout << "   50%: " << quantile(hypervolume_distribution, 0.50) << endl;
	cout << "   75%: " << quantile(hypervolume_distribution, 0.75) << endl;

	// csv export
	if (! csv_filename.empty())
	{
		ComparatorIndicesLast comparator_last(&best_points);
		vector<size_t> index(N);
		for (size_t i=0; i<N; i++) index[i] = i;
		sort(index.begin(), index.end(), comparator_last);
		ofstream ofs(csv_filename);
		for (size_t i=0; i<N; i++)
		{
			Vector2 const& v = best_points[index[i]];
			ofs << setw(16) << setprecision(10) << v[0] << ", " << setw(16) << setprecision(10) << v[1] << endl;
		}
	}

	// encapsulated postscript export
	if (! eps_filename.empty())
	{
		// find minima and maxima, but include the origin in the range
		Vector2 minimum(0.0, 0.0);
		Vector2 maximum(0.0, 0.0);
		for (size_t i=0; i<best_points.size(); i++)
		{
			Vector2 const& p = best_points[i];
			for (size_t j=0; j<2; j++)
			{
				if (p[j] < minimum[j]) minimum[j] = p[j];
				if (p[j] > maximum[j]) maximum[j] = p[j];
			}
		}

		// write header
		ofstream ofs(eps_filename);
		ofs << "%!PS-Adobe-2.0\n"
				"%%Creator: hyp3d\n"
				"%%Title: Pareto front\n"
				"%%Pages: 1\n"
				"%%PageOrder: Ascend\n"
				"%%BoundingBox: 0 0 400 400\n"
				"%%EndComments\n"
				"gsave\n"
				"1 setlinewidth [] 0 setdash 0 setgray\n"
				"(Helvetica) findfont\n"
				"dup length dict begin { 1 index /FID ne { def } { pop pop } ifelse } forall /Encoding ISOLatin1Encoding def currentdict end\n"
				"(Latin1) exch definefont pop\n"
				"(Latin1) findfont 10 scalefont setfont\n"
				"0 0 400 400 rectclip\n"
				"100 dict begin\n"
				"newpath 0 0 moveto 400 0 lineto 400 400 lineto 0 400 lineto closepath 1 setgray fill\n"
				"{\n"
				"/PopN { -1 1 { pop pop } for } bind def\n"
				"/IsNumber { dup type /integertype eq exch type /realtype eq or } bind def\n"
				"/IsString { type /stringtype eq } bind def\n"
				"/SetCol { dup IsNumber { setgray } { aload pop setrgbcolor } ifelse } bind def\n"
				"/SetFontFaceAndSize { exch findfont dup length dict begin { 1 index /FID ne { def } { pop pop } ifelse } forall /Encoding ISOLatin1Encoding def currentdict end (Latin1) exch definefont pop (Latin1) findfont exch scalefont setfont } bind def\n"
				"/UnitVector2d { dup dup mul 2 index dup mul add cvr sqrt 2 index 1 index div 4 1 roll div exch pop } bind def\n"
				"/CalcPos { 3 1 roll UnitVector2d 4 index 3 index 3 index mul add 4 index 4 index 3 index mul add 7 2 roll 5 PopN } bind def\n"
				"/ArrowHeadFilled { 3 index dup mul 3 index dup mul add cvr sqrt dup 0 eq { 7 PopN } { 5 -2 roll 2 index div 3 1 roll exch div exch 5 index 4 index 3 index mul sub 5 index 5 index 3 index mul sub newpath 8 -2 roll moveto 1 index 5 index 0.5 mul 4 index mul sub 1 index 6 index 0.5 mul 6 index mul add lineto 1 index 5 index 0.5 mul 4 index mul add 1 index 6 index 0.5 mul 6 index mul sub lineto closepath fill 6 PopN } ifelse } bind def\n"
				"/Arrow { gsave 3 index 3 index newpath exch aload pop 2 copy moveto 3 -1 roll aload pop exch 4 -1 roll sub 3 1 roll exch sub 4 index aload pop 3 index 3 index 7 index neg CalcPos lineto stroke 4 index aload pop 6 2 roll 4 2 roll ArrowHeadFilled pop pop grestore } bind def\n"
				"/Text { gsave dup 0.7 mul 8 1 roll SetFontFaceAndSize 5 -1 roll aload pop translate rotate SetCol exch dup IsString not { 256 string cvs } if exch dup (left) eq { pop 0.0 2 index -2.0 div moveto show } { dup (center) eq { pop dup stringwidth pop 2.0 div neg 2 index -2.0 div moveto show } { dup (right) eq { pop dup stringwidth pop neg 2 index -2.0 div moveto show } { ([Text] invalid alignment constant) ERROR } ifelse } ifelse } ifelse pop grestore } bind def\n"
				"/FrameCircle { newpath 1 0 0 setrgbcolor dup aload pop 3 0 360 arc fill newpath 0 setgray aload pop 3 0 360 arc stroke } bind def\n"
			;

		// draw coordinate system
		ofs << "0 setgray\n"
				"[20 20] [380 20] 10 5 Arrow\n"
				"[390 15] (f) (right) 0 0 (Helvetica) 10 Text [390 12] (1) (left) 0 0 (Helvetica) 7 Text\n"
				"[20 20] [20 380] 10 5 Arrow\n"
				"[20 390] (f) (right) 0 0 (Helvetica) 10 Text [20 387] (2) (left) 0 0 (Helvetica) 7 Text\n"
			;

		Vector2 center(20, 20);
		Vector2 axis[2];
		axis[0] = Vector2(360, 0);
		axis[1] = Vector2(0, 360);

		// draw points
		for (size_t i=0; i<best_points.size(); i++)
		{
			Vector2 const& p = best_points[i];
			Vector2 q = center;
			for (size_t j=0; j<2; j++) q += (0.9 * (p[j] - minimum[j]) / (maximum[j] - minimum[j])) * axis[j];
			ofs << "[" << q[0] << " " << q[1] << "] FrameCircle\n";
		}

		// write trailer
		ofs << "newpath 0 0 moveto 400 0 lineto 400 400 lineto 0 400 lineto closepath 0 setgray 2 setlinewidth stroke\n"
				"} stopped pop\n"
				"showpage\n"
				"end\n"
				"grestore\n"
			;
	}

	// clean up and exit
	delete ex0;
	delete ex1;
	return EXIT_SUCCESS;
}
